/**! @file UsageExample.h
	The purpose of this file is to demonstrate, in a basic way, the usage of Ornatrix core host-independent interfaces. It is
	not meant to be used for other purposes.
*/
#pragma once

// Incude Ornatrix interfaces and constructs
#include "IHair.h"
#include "IHairContainer.h"
#include "IHairRenderingPropertiesContainer.h"
#include "HairRenderingProperties.h"
#include "SurfaceDependency.h"
#include "IFunction.h"

// Utility classes, just for demonstration
#include <vector>

// Use Ornatrix namespace because all used definitions are within it
using namespace Ephere::Plugins::Ornatrix;

// Use STL namespace to make things look cleaner 
using namespace std;

//! A pretty generic structure to represent a hair vertex
struct HairVertex
{
	//! The XYZ position of the vertex in object coordinates
	float positionInObjectCoordinate[3];

	//! The UV texture coordinate of this vertex
	float textureCoordinate[2];

	//! The radius of hair strand this vertex
	float radius;
};

//! A pretty generic structure to represent a single hair strand
struct HairStrand
{
	//! Vertices making up a poly-line representing the hair strand
	vector<HairVertex> vertices;
};

/** Extract hair information from Ornatrix hair
	@param sourceHair Source hair object
	@param sourceRenderingProperties Source rendering properties object
	@param resultStrands Result strands
*/
void ExtractHairInformation( const IHair& sourceHair, const IHairRenderingPropertiesContainer& sourceRenderingProperties, vector<HairStrand>& resultStrands )
{
	const auto strandCount = sourceHair.GetStrandCount();
	resultStrands.resize( strandCount );

	// Index of the mapping channel for which to return UV coordinates. It is 0 by default but can be anything if multiple mapping channels are used and a specific
	// one is needed.
	static const unsigned MappingChannelIndex = 0;

	// Create default rendering properties to use for hair when it doesn't care about rendering properties
	HairRenderingProperties defaultRenderingProperties;
	defaultRenderingProperties.SetUsed( true );

	// Iterate over all the strands in the hair
	for( auto strandIndex = 0u; strandIndex < strandCount; strandIndex++ )
	{
		auto& resultStrand = resultStrands[strandIndex];

		const auto pointCount = sourceHair.GetStrandPointCount( strandIndex );
		resultStrand.vertices.resize( pointCount );

		// Iterate over the strand's points
		for( auto pointIndex = 0u; pointIndex < pointCount; pointIndex++ )
		{
			auto& resultVertex = resultStrand.vertices[strandIndex];

			// Get the object-space position of the point
			const auto point = sourceHair.GetStrandPointInObjectCoordinates( strandIndex, pointIndex );
			resultVertex.positionInObjectCoordinate[0] = float( point.x );
			resultVertex.positionInObjectCoordinate[1] = float( point.y );
			resultVertex.positionInObjectCoordinate[2] = float( point.z );

			// Get the texture coordinate at point
			const auto textureCoordinate = sourceHair.GetTextureCoordinate( strandIndex, pointIndex, MappingChannelIndex );
			resultVertex.textureCoordinate[0] = float( textureCoordinate.x );
			resultVertex.textureCoordinate[1] = float( textureCoordinate.y );

			// Try to find the rendering properties for the strand group to which this strand belongs
			const IHairRenderingProperties* renderingProperties = nullptr;
			if( sourceHair.KeepsSurfaceDependency() )
			{
				renderingProperties = sourceRenderingProperties.GetHairRenderingProperties( sourceHair.GetSurfaceDependency( strandIndex ).group );
			}

			// If there are no rendering properties specifically for the strand group we requested just use the default ones
			if( renderingProperties == nullptr )
			{
				renderingProperties = &defaultRenderingProperties;
			}

			if( renderingProperties->IsUsed() )
			{
				resultVertex.radius = renderingProperties->GetGlobalStrandRadius();

				// If a curve is used multiply radius by it
				if( renderingProperties->GetRadiusCurve() != nullptr )
				{
					const auto positionAlongStrand = float( pointIndex ) / float( pointCount - 1 );
					resultVertex.radius *= renderingProperties->GetRadiusCurve()->Evaluate( positionAlongStrand );
				}

				// If a thickness map is used multiply radius by it
				if( renderingProperties->GetRadiusMap() != nullptr )
				{
					// Use whichever evaluation function current host supports to evaluate the texture at current texture coordinate.
					// For example, in 3dsmax, this would involve calling Texmap::EvalMono
#if FALSE
					resultVertex.radius *= Evaluate( renderingProperties->GetRadiusMap(), textureCoordinate );
#endif
				}

				// If a strand channel is provided multiply radius by it
				if( renderingProperties->GetRadiusStrandDataChannel() != 0 )
				{
					// TODO: Provide sample code
				}
			}
			// If rendering properties were turned off
			else
			{
				resultVertex.radius = 1.0f;
			}
		}
	}
}

// Following snippets show how to extract IHair and IHairRenderingPropertiesContainer interfaces from various hosts:

#ifdef IS_COMPILING_FOR_3DSMAX
#include <max.h>

enum OrnatrixInterfaceIDs
{
	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to IHair
	IHairInterfaceID = I_USERINTERFACE + 4414,

	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to Ephere::Plugins::Ornatrix::IHair
	IHostIndependentHairInterfaceID = IHairInterfaceID + 10,

	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to IGuides
	IGuidesInterfaceID = I_USERINTERFACE + 51035,

	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to Ephere::Plugins::Ornatrix::IGuides
	IHostIndependentGuidesInterfaceID = IGuidesInterfaceID + 10,

	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to IHairRenderingPropertiesContainer
	IHairRenderingPropertiesContainerInterfaceID = I_USERINTERFACE + 4415,

	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to IHairRenderingPropertiesContainer
	IHostIndependentHairRenderingPropertiesContainerInterfaceID = IHairRenderingPropertiesContainerInterfaceID + 10
};

/** Gets Ornatrix hair interface and rendering properties from a 3dsmax object.
	@return true if object was Ornatrix hair, false otherwise
*/
bool GetHairInterfaces( Object& maxObject, IHair*& hairInterface, IHairRenderingPropertiesContainer*& renderingProperties )
{
	hairInterface = static_cast<IHair&>( maxObject.GetInterface( IHostIndependentHairInterfaceID ) );
	renderingProperties = static_cast<IHairRenderingPropertiesContainer&>( maxObject.GetInterface( IHostIndependentHairRenderingPropertiesContainerInterfaceID ) );

	return hairInterface != nullptr && renderingProperties != nullptr;
}

/** Gets Ornatrix hair interface and rendering properties from a 3dsmax scene node.
	@return true if object was Ornatrix hair, false otherwise
*/
bool GetHairInterfaces( INode& maxNode, IHair*& hairInterface, IHairRenderingPropertiesContainer*& renderingProperties )
{
	return GetHairInterfaces( maxNode.EvalWorldState( 0 ).obj, hairInterface, renderingProperties );
}
#elif defined( IS_COMPILING_FOR_MAYA )
#include <maya/MDataHandle.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MPlug.h>
#include <maya/MPxData.h>

bool GetHairInterfaces( const MObject& hairShapeNode, IHair*& hairInterface, IHairRenderingPropertiesContainer*& renderingProperties )
{
	MFnDependencyNode fn( hairShapeNode );

	MStatus status;

	// Get hair attribute
	// NOTE: If using GuideShape instead of HairShape this method would look identical, except instead of "outputHair" we would use "outputGuides" for the plug name below
	auto hairPlug = fn.findPlug( "outputHair", &status );
	if( !status )
	{
		return false;
	}

	auto renderSettingsPlug = fn.findPlug( "outputRenderSettings", &status );
	if( !status )
	{
		return false;
	}

	// We can cast the data handle to a hair container, from which we can get the hair interface
	auto hairDataHandle = hairPlug.asMDataHandle();
	hairInterface = dynamic_cast<IHairContainer*>( hairDataHandle.asPluginData() )->GetHair();

	// We can cast render settings data directly to IHairRenderingPropertiesContainer
	auto renderSettingsDataHandle = renderSettingsPlug.asMDataHandle();
	renderingProperties = dynamic_cast<IHairRenderingPropertiesContainer*>( renderSettingsDataHandle.asPluginData() );

	return hairInterface != nullptr && renderingProperties != nullptr;
}
#endif